///////////////////////////////////////////////////////////////////////////////
// NABLA - a Numerical Analysis Based LAnguage                               //
//                                                                           //
// Copyright (C) 2014~2017 CEA/DAM/DIF                                       //
// IDDN.FR.001.520002.000.S.P.2014.000.10500                                 //
//                                                                           //
// Contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr          //
//                                                                           //
// This software is a computer program whose purpose is to translate         //
// numerical-analysis specific sources and to generate optimized code        //
// for different targets and architectures.                                  //
//                                                                           //
// This software is governed by the CeCILL license under French law and      //
// abiding by the rules of distribution of free software. You can  use,      //
// modify and/or redistribute the software under the terms of the CeCILL     //
// license as circulated by CEA, CNRS and INRIA at the following URL:        //
// "http://www.cecill.info".                                                 //
//                                                                           //
// The CeCILL is a free software license, explicitly compatible with         //
// the GNU GPL.                                                              //
//                                                                           //
// As a counterpart to the access to the source code and rights to copy,     //
// modify and redistribute granted by the license, users are provided only   //
// with a limited warranty and the software's author, the holder of the      //
// economic rights, and the successive licensors have only limited liability.//
//                                                                           //
// In this respect, the user's attention is drawn to the risks associated    //
// with loading, using, modifying and/or developing or reproducing the       //
// software by the user in light of its specific status of free software,    //
// that may mean that it is complicated to manipulate, and that also         //
// therefore means that it is reserved for developers and experienced        //
// professionals having in-depth computer knowledge. Users are therefore     //
// encouraged to load and test the software's suitability as regards their   //
// requirements in conditions enabling the security of their systems and/or  //
// data to be ensured and, more generally, to use and operate it in the      //
// same conditions as regards security.                                      //
//                                                                           //
// The fact that you are presently reading this means that you have had      //
// knowledge of the CeCILL license and that you accept its terms.            //
//                                                                           //
// See the LICENSE file for details.                                         //
///////////////////////////////////////////////////////////////////////////////
with ℝ²;
#include "assert.h"

// Cette version serait à comparer avec tests/ns4n/glcNxtStp2.n

// ****************************************************************************
// * Options du schéma Glace Euler/Sans reconstruction
// ****************************************************************************
options{
  // Mesh options
  ℝ LENGTH                    = 1.125;
  ℕ X_EDGE_ELEMS              = 8;
  ℕ Y_EDGE_ELEMS              = 8;
  ℕ Z_EDGE_ELEMS              = 1;
  Bool DEBUG = false;
  Bool TRACE = false;
  Bool option_test_sod = true;
      
  Real γ = 1.4;
  Real option_u_ini = 0.0;
  Real option_ini_zg_ρ = 1.0;
  Real option_ini_zd_ρ = 0.125;
  Real option_ini_zg_p = 1.0; 
  Real option_ini_zd_p = 0.1;

  Real option_x_min = 0.0;
  Real option_y_min = 0.0;
  Real option_x_interface = 0.5625;
  Real option_x_max = 1.125;
  Real option_y_max = 1.125;
  
  Real option_δt_initial = 0.0001;         // ! utilisé en dur dans le backend lambda !
  Real option_δt_min = 1e-12;
  Real option_δt_max = 0.01;
  Real option_stoptime = 0.1;              // ! utilisé en dur dans le backend lambda !
  Real option_δt_control = 0.1;            // Contrôle du pas de temps hydrodynamique
  Real option_δt_min_variation = 0.99;     // Décroissance maximale tolérée du pas de temps
  Real option_δt_max_variation = 0.10;     // Amplitude maximale de croissance du pas de temps
  Bool option_δt_last_exact = true;
  
  Integer option_max_iterations = 32768;   // ! utilisé en dur dans le backend lambda !
};

// ****************************************************************************
// * Déclaration des variables aux noeuds
// ****************************************************************************
nodes{
  Real3   node_u;                          // Vitesse nodale du maillage
  Real3x3 node_u_matrix;                   // Matrice aux sommets pour le calcul de la vitesse
  Real3   node_u_second_member;            // Second membre aux sommets pour le calcul de la vitesse
};

// ****************************************************************************
// * Déclaration des variables aux mailles
// ****************************************************************************
cells{  
  Real m;                                  // Masse
  Real V;                                  // Volume des mailles
  Real ρ;                                  // Densités
  Real p;                                  // Pression
  Real3 u;                                 // Vitesse moyenne
  Real css;
  Real glace_δt;
  Real absCQs[nodes];
  Real3 CQs[nodes];                        // Résultantes aux sommets
  Real3x3 AQs[nodes];                      // Matrices de contributions aux sommets
  Real ρΕ;                                 // Énergie massique
  Real ιΕ;                                 // Énergie interne
  Real3 reconstructed_u[nodes];            // Tableau des vitesses reconstruites
  Real reconstructed_p[nodes];             // Tableau des pressions reconstruites
  Real3 momentum_fluxes_Σ;
  Real total_energy_fluxes_Σ;
};

// ****************************************************************************
// * Globals
// ****************************************************************************
global{
  ℝ global_δt;
};

// ****************************************************************************
// * Partie boucle de calcul ]+0.0,+∞[ 
// ****************************************************************************

// ****************************************************************************
// * computeAQsMatrices
// ****************************************************************************
∀ cells @ 9.0 {
  const Real ρc = ρ*css;
  ∀ node{
    //const int k = xs_cell_node[c+NABLA_NB_CELLS*n];
    //if (TRACE) printf("\n[32m[std] k: %d %d %d %d[0m",k0,k1,k2,k3);
    AQs = CQs⊗CQs;
    AQs *= ρc/absCQs;
    //AQs = 10*c+n+Real3x3(Real3(0.11,0.22,0.33),Real3(0.44,0.55,0.66),Real3(0.77,0.88,0.99));
    if (DEBUG) printf("\n\t[32m[std] AQs=");
    if (DEBUG) cout<<AQs<<"[0m";
  }
}

// ****************************************************************************
// * temporalSchemeReconstructPressure
// ****************************************************************************
∀ cells @ 10.0{
  ∀ node{
    reconstructed_p=p;
    if (DEBUG)
      printf("\n\t[32m[std] reconstructed_p=%.15e[0m",
             reconstructed_p);
  }
}

// ****************************************************************************
// * temporalSchemeReconstructVelocity
// ****************************************************************************
∀ cells @ 10.0{
  ∀ node{
    reconstructed_u=u;
    if (DEBUG)
      printf("\n\t[32m[std] reconstructed_u=(%.15e,%.15e,%.15e)[0m",
           reconstructed_u.x,reconstructed_u.y,reconstructed_u.z);
  }
}

// ****************************************************************************
// * temporalSchemeIniMatricesAndSecondMember
// ****************************************************************************
∀ nodes @ 11.0{
  node_u_matrix=Real3x3();
  node_u_matrix.z=Real3(0,0,1);
  node_u_second_member=0;
}

// ****************************************************************************
// * temporalSchemeAssembleMatricesAndSecondMember
// ****************************************************************************
∀ cells @ 12.0{
  //if (TRACE) printf("\n\n[37m[temporalSchemeAssembleMatricesAndSecondMember] Cell #%d[0m",uid);
  //const int kCell=uid;
  ∀ node{
    //const int k = xs_cell_node[n*NABLA_NB_CELLS+kCell];
    //if (TRACE) printf("\n[37m[temporalSchemeAssembleMatricesAndSecondMember] k=%d[0m",n);
    const Real3x3 local_matrix = AQs;
    //if (DEBUG) printf("\n\t[32m[std] AQs=");
    //if (DEBUG) cout<<AQs<<"[0m";
    node_u_matrix += local_matrix;
    /*if (DEBUG)
    printf("\n\t[31m[temporalSchemeAssembleMatricesAndSecondMember] node_u_matrix=\n\t\t[\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]][0m",
           node_u_matrix.x.x,node_u_matrix.x.y,node_u_matrix.x.z,
           node_u_matrix.y.x,node_u_matrix.y.y,node_u_matrix.y.z,
           node_u_matrix.z.x,node_u_matrix.z.y,node_u_matrix.z.z);*/
    node_u_second_member += local_matrix ⨂ reconstructed_u;
    node_u_second_member += CQs*reconstructed_p;
  }
}
∀ nodes @ 12.1{ // On revient APRES les écritures aux noeuds pour dumper
  if (DEBUG)
    printf("\n\t[31m[std] Before BOUNDARIES node_u_matrix[%d]=\n\t\t[\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]\n\t\t\
[%.15e %.15e %.15e]][0m",uid,
           node_u_matrix.x.x,node_u_matrix.x.y,node_u_matrix.x.z,
           node_u_matrix.y.x,node_u_matrix.y.y,node_u_matrix.y.z,
           node_u_matrix.z.x,node_u_matrix.z.y,node_u_matrix.z.z);
  if (DEBUG)
    printf("\n\t[31m[std] Before BOUNDARIES node_u_second_member[%d]=\
(%.15e,%.15e,%.15e)[0m",uid,
           node_u_second_member.x,
           node_u_second_member.y,
           node_u_second_member.z);
}

// ****************************************************************************
// * Le calcul aux boundaries se fait ici @ 13.1,13.2
// ****************************************************************************

// ****************************************************************************
// * synchronize pour Arcane node_u_matrix & node_u_matrix_second_member
// ****************************************************************************
void synchronizeNodeUMatrixAndSecondMember(void)@ 13.4{
  synchronize(node_u_matrix);
  synchronize(node_u_second_member);
}

// ****************************************************************************
// * temporalSchemeComputeNodesVelocity
// ****************************************************************************
∀ nodes @ 14.0{
  if (DEBUG)
    printf("\n\t[31m[std] node_u_matrix=[0m\n\t\t[31m[\
[%.15e %.15e %.15e]\n\t\t[%.15e %.15e %.15e]\n\t\t[%.15e %.15e %.15e]][0m",
         node_u_matrix.x.x,node_u_matrix.x.y,node_u_matrix.x.z,
         node_u_matrix.y.x,node_u_matrix.y.y,node_u_matrix.y.z,
         node_u_matrix.z.x,node_u_matrix.z.y,node_u_matrix.z.z);
  const Real determinent = matrixDeterminant(node_u_matrix);
  if (DEBUG)
    printf("\n\t[31m[std] node_u_second_member=(%.15e,%.15e,%.15e)[0m",
         node_u_second_member.x,
         node_u_second_member.y,
         node_u_second_member.z);
  assert(determinent!=0.0);
  node_u_matrix = inverseMatrix(node_u_matrix, determinent);
  node_u = node_u_matrix ⨂ node_u_second_member;
  node_u.z=0.0;
  if (DEBUG)
    printf("\n\t[31m[std] node_u[%d]=(%.15e,%.15e,%.15e)[0m",
           uid,node_u.x,node_u.y,node_u.z);
}
  
// ****************************************************************************
// * temporalSchemeIniFluxesSum
// ****************************************************************************
∀ cells @ 15.0{
  momentum_fluxes_Σ=total_energy_fluxes_Σ=0.0;
}

// ****************************************************************************
// * temporalSchemeComputeStandardFluxesSum
// ****************************************************************************
∀ cells @ 16.0{
  ∀ node{
    const Real3 Δu = reconstructed_u-node_u;
    Real3 FQs = AQs ⨂ Δu;
    FQs += reconstructed_p*CQs;
    momentum_fluxes_Σ -= FQs;
    total_energy_fluxes_Σ -= FQs⋅node_u;
  }
}

// ****************************************************************************
// * temporalSchemeApplyFluxesStandard
// ****************************************************************************
∀ cells @ 19.0{
  if (DEBUG) printf("\n\t[32m[n+1] deltaT=%.15e[0m",δt);
  if (DEBUG) printf("\n\t[32m[n+1] pre u=(%.15e,%.15e,%.15e)[0m",u.x,u.y,u.z);
  if (DEBUG)
    printf("\n\t[32m[n+1] momentum_fluxes_Σ=(%.15e,%.15e,%.15e)[0m",
           momentum_fluxes_Σ.x,
           momentum_fluxes_Σ.y,
           momentum_fluxes_Σ.z);
  u += δt*momentum_fluxes_Σ/m;
  if (DEBUG)
    printf("\n\t[32m[n+1] post u=(%.15e,%.15e,%.15e)[0m",u.x,u.y,u.z);
  ρΕ += δt*total_energy_fluxes_Σ/m;
}

// ****************************************************************************
// * temporalSchemeMoveNodes
// ****************************************************************************
∀ nodes @ 22.0{
  if (DEBUG)
    printf("\n\t[32m[n+1] greek_deltat=%.15e[0m",δt);
  coord += δt*node_u;
  if (DEBUG)
    printf("\n\t[31m[n+1] node_coord[%d]=(%.15e,%.15e,%.15e)[0m",
         uid, coord.x, coord.y, coord.z);
}

// ****************************************************************************
// * geomComputeCQsQuad @ 23.0
// * geomComputeSurfaceAndVolume @ 24.0 
// ****************************************************************************

// ****************************************************************************
// * glaceEOS vient se placer ici: [26.0 → 30.0]
// ****************************************************************************

// ****************************************************************************
// * Réduction min vers le global_δt
// ****************************************************************************
∀ cells global_δt <?= glace_δt @ 31.0;

// ****************************************************************************
// * glaceTimeStepCtrl
// ****************************************************************************
void hydroTimeStepCtrl(void) @ 35.0{
  Real new_δt=option_δt_control*global_δt;
  if (DEBUG)
    printf("\n[1;33m option_δt_control=%.15e, global_δt=%.15e[0m",
           option_δt_control,global_δt);
  // on calcule les variations min et max autorisées
  const Real max_time_step = (1.0+option_δt_max_variation)*δt;
  const Real min_time_step = (1.0-option_δt_min_variation)*δt;
  if (new_δt > max_time_step) new_δt=max_time_step;
  if (new_δt < min_time_step) new_δt=min_time_step;
  new_δt = min(new_δt, option_δt_max);
  new_δt = max(new_δt, option_δt_min);
  // le dernier calcul se fait exactement au temps final
  if (option_δt_last_exact && ((time + new_δt) > option_stoptime)) 
      new_δt = option_stoptime - time;
  if ((time>=option_stoptime) || (new_δt==0.0)) exit;
  δt=new_δt;
  if (DEBUG)
    printf("\n[1;33m final δt=%.15e[0m\n",δt);
  if (option_max_iterations!=0 && iteration >= option_max_iterations) exit;
}
