///////////////////////////////////////////////////////////////////////////////
// This NABLA port is an implementation of the PENNANT Mini-App              //
// from Charles R. Ferenbaugh (cferenba@lanl.gov), LANL                      //
// ∇ contributor(s): CAMIER Jean-Sylvain - Jean-Sylvain.Camier@cea.fr        //
///////////////////////////////////////////////////////////////////////////////
// PENNANT is an unstructured mesh physics mini-app designed for advanced    //
// architecture research. It contains mesh data structures and a few physics //
// algorithms adapted from the LANL rad-hydro code FLAG, and gives a sample  //
// of the typical memory access patterns of FLAG.                            //
///////////////////////////////////////////////////////////////////////////////
// Copyright 2012, Los Alamos National Security, LLC. All rights reserved.   //
// This software was produced under US Government contract DE-AC52-06NA25396 //
// for Los Alamos National Laboratory, which is operated by Los Alamos       //
// National Security, LLC for the U.S. Department of Energy.                 //
// The US Government has rights to use, reproduce and distribute this        //
// software.                                                                 //
// NEITHER THE GOVERNMENT NOR LOS ALAMOS NATIONAL SECURITY, LLC MAKES ANY    //
// WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LIABILITY FOR THE USE OF     //
// THIS SOFTWARE.                                                            //
// If this software is modified to produce derivative works, such modified   //
// software should be clearly marked, so as not to confuse it with the       //
// version available from LANL.                                              //
// Additionally, redistribution and use in source and binary forms, with or  //
// without modification, are permitted provided that the following           //
// conditions are met:                                                       //
//  1. Redistributions of source code must retain the above copyright notice,//
//     this list of conditions and the following disclaimer.                 //
//  2. Redistributions in binary form must reproduce the above copyright     //
//     notice, this list of conditions and the following disclaimer in the   //
//     documentation and/or other materials provided with the distribution.  //
//  3. Neither the name of Los Alamos National Security, LLC, Los Alamos     //
//     National Laboratory, LANL, the U.S. Government, nor the names of its  //
//     contributors may be used to endorse or promote products derived from  //
//     this software without specific prior written permission.              //
///////////////////////////////////////////////////////////////////////////////
// THIS SOFTWARE IS PROVIDED BY LOS ALAMOS NATIONAL SECURITY, LLC AND        //
// CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,    //
// BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS //
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED.                                  //
// IN NO EVENT SHALL LOS ALAMOS NATIONAL SECURITY, LLC OR CONTRIBUTORS BE    //
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR       //
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF      //
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS  //
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN   //
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)   //
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF    //
// THE POSSIBILITY OF SUCH DAMAGE.                                           //
///////////////////////////////////////////////////////////////////////////////
with ℝ²;

// ****************************************************************************
// Default options should match test/sedovsmall/sedovsmall.pnt
// cstop=100, tstop=1.0, meshtype=rect, meshparams=3 3 1.125 1.125
// subregion=0.0 0.3 0.0 0.3, rinitsub=1.0, einitsub=40.222, bcx=0.0 1.125
// bcy=0.0 1.125, ssmin=0.1, q1=0.1, q2=1.0, dtinit=0.0025
// ****************************************************************************
options{
  // Mesh options
  ℝ LENGTH       = 1.125;
  ℕ X_EDGE_ELEMS = 3;
  ℕ Y_EDGE_ELEMS = 3;
  ℕ Z_EDGE_ELEMS = 1;
  ℕ option_max_iterations = 0;// ! utilisé en dur dans le backend lambda !
  ℝ option_δt_initial = 1e-7; // ! utilisé en dur dans le backend lambda !
  ℝ option_stoptime = 1.0;    // ! utilisé en dur dans le backend lambda !
  // Pennant options
  ℾ trace       = false;
  ℕ cstop       = 100;    // simulation stop cycle
  ℝ tstop       = 1.0;    // simulation stop time
  ℝ dtmax       = 1.e99;  // maximum timestep size
  ℝ dtinit      = 0.0025; // initial timestep size
  ℝ dtfac       = 1.2;    // factor limiting timestep growth
  ℕ dtreport    = 10;     // frequency for timestep reports
  ℝ cfl         = 0.6;    // Courant number, limits timestep
  ℝ cflv        = 0.1;    // volume change limit for timestep
  ℝ rinit       = 1.0;    // initial density for main mesh
  ℝ einit       = 0.0;    // initial energy for main mesh
  ℾ subregion   = true;
  ℝ subrgn_xmin = 0.0;
  ℝ subrgn_xmax = 0.3;
  ℝ subrgn_ymin = 0.0;
  ℝ subrgn_ymax = 0.3;
  ℝ rinitsub    = 1.0;    // initial density in subregion
  ℝ einitsub    = 40.222; // initial energy in subregion
  ℝ uinitradial = 0.0;    // initial velocity in radial direction
  ℝ ggamma      = 5./3.;  // coeff. for ideal gas equation
  ℝ ssmin       = 0.1;    // minimum sound speed for gas
  ℝ alpha       = 0.5;    // alpha coefficient for TTS model
  ℝ qgamma      = 5./3.;  // gamma coefficient for Q model
  ℝ q1          = 0.1;    // linear coefficients for Q model
  ℝ q2          = 1.0;    // quadratic coefficients for Q model
  ℾ chaos       = false;
  ℝ chaos_seed  = 1.1234567890123;
};


// ****************************************************************************
// * Trace & Debug
// ****************************************************************************
enum{red=31,green,yellow,blue,magenta,cyan,white};
#define dbg(...) if (trace) printf("%s%s%s","\n[35m",__VA_ARGS__,"[m")
#define dbc(clr,...) if (trace) cout<<"["<<clr<<"m"<<__VA_ARGS__<<"[m"<<endl;


// ****************************************************************************
// * Zone/Sides/Corners Variables
// ****************************************************************************
cells{
  ℝ² zx;    // zone center coordinates
  ℝ² zxp;   // zone ctr coords, middle of cycle
  ℝ zarea;  // zone area
  ℝ zvol;   // zone volume
  ℝ zareap; // zone area, middle of cycle
  ℝ zvolp;  // zone volume, middle of cycle
  ℝ zvol0;  // zone volume, start of cycle
  ℝ zdl;    // zone characteristic length
  ℝ zm;     // zone mass
  ℝ zr;     // zone density
  ℝ zrp;    // zone density, middle of cycle
  ℝ ze;     // zone specific internal energy (energy per unit mass)
  ℝ zetot;  // zone total internal energy
  ℝ zw;     // zone work done in cycle
  ℝ zwrate; // zone work rate
  ℝ zp;     // zone pressure
  ℝ zss;    // zone sound speed
  ℝ zdu;    // zone velocity difference
  ℝ z0per;  // zone tmp variable in PolyGas calcStateAtHalf
  ℝ² z0uc;  // zone tmp centered velocity
  // Sides
  ℝ sarea[faces];  // area
  ℝ svol[faces];   // volume
  ℝ sareap[faces]; // area, middle of cycle
  ℝ svolp[faces];  // volume, middle of cycle
  ℝ² ssurfp[faces];// surface vector
  ℝ smf[faces];    // mass fraction
  ℝ² sfp[faces];   // force from pressure
  ℝ² sfq[faces];   // force from artificial visc.
  ℝ² sft[faces];   // force from tts
  // Corners
  ℝ² cftot[nodes]; // force, total from all sources
  ℝ cmaswt[nodes]; // contribution to pmaswt
  ℝ c0area[nodes]; // tmp
  ℝ c0evol[nodes]; // tmp
  ℝ c0du[nodes];   // tmp
  ℝ c0div[nodes];  // tmp
  ℝ c0cos[nodes];  // tmp
  ℝ² c0qe0[nodes]; // tmp
  ℝ² c0qe1[nodes]; // tmp
  ℝ c0w[nodes];    // tmp
  ℝ c0rmu[nodes];  // tmp
};


// ****************************************************************************
// * Point Variables
// ****************************************************************************
nodes{
  ℝ² px,px0,pxp; // point coordinates, start & middle of cycle
  ℝ² pu,pu0;     // point velocity
  ℝ² pap,pf;     // point acceleration, force
  ℝ pmaswt;      // point mass, weighted by 1/r
};


// ****************************************************************************
// * Edge Variables
// ****************************************************************************
faces{
  ℝ² ex,exp; // edge center coordinates & middle of cycle
  ℝ elen;    // edge length
};


// ****************************************************************************
// * Global Variables
// ****************************************************************************
global{
  ℕ cycle;  // simulation cycle number
  ℝ dt;     // current timestep
  ℝ dtrec;  // maximum timestep for hydro
  ℝ dvovmax;//
};


// ****************************************************************************
// * Partie d'initialisation ]-∞,-0.0[
// ****************************************************************************
iniCout @ -48.0 {cout<<"\n[m";}

∀ nodes iniMeshCoords @ -32.0{
  px = coord;
  px *= (chaos)?chaos_seed:1.0;
  dbc(yellow,"[iniMeshCoords] px["<<n<<"]="<<px);
}

∀ cells iniMeshCalcCtrs_zx @ -31.9 {
  zx=0;
  ∀ node zx+=px;
  zx /= nbNode;
  dbc(yellow,"[Mesh::calcCtrs] zx["<<c<<"]="<<zx);
}

∀ faces iniMeshCalcCtrs_ex @ -31.88 {
  ex = 0.0;
  ∀ node ex += px; 
  ex /= nbNode;
  assert(nbNode==2);
  //dbc(yellow,"[Mesh::calcCtrs] ex["<<uid<<"]="<<ex);
}

∀ cells iniMeshCalcVols @ -31.8 {
  const int N = nbNode;
  ℝ³ X[4]; ∀ node X[#]=px;
  zvol=0;
  zarea=0; 
  ∀ node{
    const int i=#;
    const int j=(#+1)%N;
    sarea = ½ * cross2D(X[j]-X[i],zx-X[i]);
    svol = ⅓* sarea * (X[i].x+X[j].x+zx.x);
    zarea += sarea;
    zvol  += svol;
  }
  dbc(yellow,"[iniMeshCalcVols] zvol="<<zvol);
}

∀ cells iniMeshCalcSideFracs @ -31.7 {
  ∀ face {
    smf = sarea / zarea;
    //dbc(yellow,"[iniMeshCalcSideFracs] smf["<<c<<"]="<<smf);
  }
}

iniTstOptions @ -30.0 {
  dbc(magenta,"optTest");
  if (cstop == 999999 && tstop == 1.e99) {
    cerr << endl << "Must specify either cstop or tstop" << endl;
    exit;
  }
}

iniGlobals @ -20.0 {
  dbc(magenta,"iniGlobals");
  time = 0.0;
  cycle = 0;
}

∀ cells iniZr @ -14.9 { zr=rinit; }
∀ cells iniZe @ -14.9 { ze=einit; }
∀ cells iniZwrate @ -14.9 { zwrate=0.0; }
∀ cells @ -14.8 if (subregion) {
  const double eps = 1.e-12;
  if (!(zx.x > (subrgn_xmin - eps) && zx.x < (subrgn_xmax + eps) &&
        zx.y > (subrgn_ymin - eps) && zx.y < (subrgn_ymax + eps))) continue;
  dbc(white,"[subregion] zx["<<c<<"] in subregion!");
  zr = rinitsub;
  ze = einitsub;
}

∀ cells iniZm @ -14.7 {
  zm = zr * zvol;
  dbc(white,"[Hydro::init] zm["<<c<<"]="<<zm);
}

∀ cells iniZetot @ -14.6 {
  zetot = ze * zm;
  dbc(white,"[Hydro::init] zetot["<<c<<"]="<<zetot);
}

∀ nodes iniPuRadial @ -14.5 if (uinitradial!=0.0) {
  const ℝ eps = 1.e-12;
  const ℝ pmag = length(px);
  if (pmag > eps) pu = uinitradial * px / pmag;
  else pu = 0.0;
}
∀ nodes iniPuStd @ -14.5 if (uinitradial==0.0) { pu=0.0; }

iniRstDtHydro @ -14.4 { dtrec = 1.e99; }


// ****************************************************************************
// * Partie boucle de calcul ]+0.0,+∞[ 
// ****************************************************************************

incCycle @ 1.0 { cycle += 1; }

calcGlobalDt @ 2.0 {
  // Save timestep from last cycle
  const ℝ dtlast = dt;
  // Compute timestep for this cycle
  dt = dtmax;
  if (cycle == 1) {    // compare to initial timestep
    dbc(magenta,"[Driver::calcGlobalDt] initial timestep, dtinit="<<dtinit);
    if (dtinit < dt) dt = dtinit;
  } else {
    const ℝ dtrecover = dtfac * dtlast;
    if (dtrecover < dt) dt = dtrecover;
  }
  // compare to time-to-end
  if ((tstop - time) < dt) dt = tstop - time;
  // compare to hydro dt
  if (dtrec < dt) dt = dtrec;
  dbc(magenta,"[calcGlobalDt] dt="<<dt);
}

// Hydro::doCycle
// Begin hydro cycle

∀ nodes saveOffPxFromPreviousCycle @ 3.0 { px0=px; }
∀ nodes saveOffPuFromPreviousCycle @ 3.0 { pu0=pu; }
∀ cells saveOffZvolFromPreviousCycle @ 3.0 { zvol0=zvol; }

// ===== Predictor step =====
// 1. advance mesh to center of time step
∀ nodes advPosHalf0 @ 3.1 {
  pxp = px0 + pu0 * ½ * dt;
  dbc(white,"[Hydro::advPosHalf] pxp"<<pxp);
}

// 1a. compute new mesh geometry
∀ cells computeZxp @ 3.2 {
  zxp=0;
  ∀ node zxp+=pxp;
  zxp/= nbNode;
  dbc(yellow,"[Mesh::calcCtrs] zxp["<<c<<"]="<<zxp);
}

∀ faces computeExp @ 3.2 {
  exp = 0.0;
  ∀ node exp += pxp;
  exp /= nbNode;
  //dbc(yellow,"[Mesh::calcCtrs] exp["<<f<<"]="<<exp);
}

∀ cells calcVols @ 3.4 {
  const int N = nbNode;
  ℝ³ X[4]; ∀ node X[#]=pxp;
  zareap=0; 
  zvolp=0;
  ∀ node {
    const ℕ j=(#+1)%N;
    sareap = ½ * cross2D(X[j]-X[#],zxp-X[#]);
    svolp = ⅓* sareap * (X[#].x+X[j].x+zxp.x);
    zareap += sareap;
    zvolp  += svolp;
  } 
  dbc(yellow,"[calcVols] zvolp="<<zvolp);
}
//∀ cells dumpZareap @ 3.41{dbc(yellow,"[calcVols] zareap="<<zareap);} 
//∀ cells dumpSareap @ 3.42{ ∀ node dbc(yellow,"[calcVols] sareap="<<sareap);} 

∀ cells calcSurfVecs @ 3.5 {
  ∀ face{
    ssurfp = rotateCCW(exp - zxp);
    //dbc(yellow,"[calcSurfVecs] ssurfp="<<ssurfp);
  }
}

∀ faces calcEdgeLen @ 3.5 {
  elen = length(pxp[1]-pxp[0]);
  //dbc(yellow,"[calcEdgeLen] elen["<<f<<"]="<<elen);
}

∀ cells calcCharLen @ 3.6 {
  zdl = 1.e99;
  ∀ face {
    const ℝ sdl = 4.0 * sareap / elen;
    zdl = min(zdl, sdl);
  }
  dbc(yellow,"[calcCharLen] zdl="<<zdl);
}

// 2. compute point masses
∀ cells calcRhoOne @ 4.0 {
  zrp = zm / zvolp;
  dbc(white,"[calcRhoOne] zr="<<zrp);
}

∀ cells calcCrnrMass @ 4.1 {
  const ℕ N = nbNode;
  ℝ m[4]; ∀ node m[#]=smf;
  ∀ node {
    const ℕ j=(#+N-1)%N;
    cmaswt = ½ * zrp * zareap * (m[#] + m[j]);
    dbc(white,"[calcCrnrMass] cmaswt="<<cmaswt);
  }
}

// 3. compute material state (half-advanced)
// calcStateAtHalf
∀ cells calcStateAtHalf_calcEOS @ 5.0 {
  const ℝ gm1 = ggamma - 1.0;
  const ℝ ss2 = fmax(ssmin * ssmin, 1.e-99);
  const ℝ rx = zr;
  const ℝ local_ex = fmax(ze, 0.0);
  const ℝ local_px = gm1 * rx * local_ex;
  const ℝ prex = gm1 * local_ex;
  const ℝ perx = gm1 * rx;
  const ℝ csqd = fmax(ss2, prex + perx * local_px / (rx * rx));
  zp =local_px;
  z0per = perx;
  zss = √(csqd);
}
//∀ cells dumpZss @ 5.01 {dbc(cyan,"[PolyGas::calcEOS] zss="<<zss);}
∀ cells calcStateAtHalf_next @ 5.1 {
  const ℝ dth = ½ * dt;
  const ℝ zminv = 1.0 / zm;
  const ℝ dv = (zvolp - zvol0) * zminv;
  const ℝ bulk = zr * zss * zss;
  const ℝ denom = 1.0 + ½ * z0per * dv;
  const ℝ src = zwrate * dth * zminv;
  zp += (z0per * src - zr * bulk * dv) / denom;
  //dbc(cyan,"[PolyGas::calcStateAtHalf_next] zp="<<zp);
}

// 4. compute forces
∀ cells pgasCalcForce @ 6.0 {
  ∀ face {
    sfp=-zp * ssurfp;
    dbc(cyan,"[PolyGas::calcForce] sf["<<f<<"]="<<sfp);
  }
}
∀ cells ttsCalcForce @ 6.1 {
  ∀ face {
    const ℝ svfacinv = zareap / sareap;
    const ℝ srho = zrp * smf * svfacinv;
    ℝ sstmp = fmax(zss, ssmin);
    sstmp = alpha * sstmp * sstmp;
    const ℝ sdp = sstmp * (srho - zrp);
    const ℝ³ sqq = -sdp * ssurfp;
    sft = sqq;
    dbc(magenta,"[TTS::calcForce] sft="<<sft);
  }
}

∀ cells qcsCalcForce_zoneCenteredVelocity @ 6.2 {
  z0uc=0;
  ∀ node z0uc += pu;
  z0uc /= nbNode;  
  //dbc(green,"[QCS::setCornerDiv] z0uc="<<z0uc);
}

∀ cells qcsCalcForce_setCornerDiv @ 6.3 {
  const ℕ N = nbNode;
  const ℝ³ _pu[4] = {pu[0],pu[1],pu[2],pu[3]};
  const ℝ³ _ex[4] = {exp[0],exp[1],exp[2],exp[3]};
  const ℝ _elen[4] = {elen[0],elen[1],elen[2],elen[3]};
  ∀ node {
    const ℕ i=(#+N-1)%N; // previous
    const ℕ j=(#);       // current
    const ℕ k=(#+1)%N;   // next
    //dbc(green,"[QCS::setCornerDiv] i="<<i<<", j="<<j<<", k="<<k);
    // Velocities and positions
    // 0 = point p
    const ℝ³ up0 = pu;
    const ℝ³ xp0 = pxp;
    //dbc(green,"[QCS::setCornerDiv] up0="<<up0<<", xp0="<<xp0);
    // 1 = edge e2
    const ℝ³ up1 = ½ * (_pu[#] + _pu[k]);
    const ℝ³ xp1 = _ex[j];
    //dbc(green,"[QCS::setCornerDiv] up1="<<up1<<", xp1="<<xp1);
    // 2 = zone center z
    const ℝ³ up2 = z0uc;
    const ℝ³ xp2 = zxp;
    //dbc(green,"[QCS::setCornerDiv] up2="<<up2<<", xp2="<<xp2);
    // 3 = edge e1
    const ℝ³ up3 = ½ * (_pu[i] + _pu[#]);
    const ℝ³ xp3 = _ex[i];
    //dbc(green,"[QCS::setCornerDiv] up3="<<up3<<", xp3="<<xp3);
    // compute 2d cartesian volume of corner
    const ℝ cvolume = ½ * cross2D(xp2-xp0, xp3-xp1);
    c0area = cvolume;
    //dbc(green,"[QCS::setCornerDiv] c0area="<<c0area);
    // compute cosine angle
    const ℝ³ v1 = xp3-xp0;
    const ℝ³ v2 = xp1-xp0;
    const ℝ de1 = _elen[i];
    const ℝ de2 = _elen[#];
    ℝ minelen = min(de1, de2);
    //dbc(green,"[QCS::setCornerDiv] minelen="<<minelen);
    c0cos = ((minelen < 1.e-12) ?
                 0.0 : 4.0 * dot2D(v1,v2)/(de1*de2));
    //dbc(green,"[QCS::setCornerDiv] c0cos="<<c0cos);
    // compute divergence of corner
    c0div = (cross2D(up2-up0, xp3-xp1)-cross2D(up3-up1, xp2-xp0))/(2.0*cvolume);
    //dbc(green,"[QCS::setCornerDiv] c0div="<<c0div);
    // compute evolution factor
    const ℝ³ dxx1 = ½ * (xp1+xp2-xp0-xp3);
    const ℝ³ dxx2 = ½ * (xp2+xp3-xp0-xp1);
    const ℝ dx1 = length(dxx1);
    const ℝ dx2 = length(dxx2);
    // average corner-centered velocity
    const ℝ³ duav = 0.25 * (up0+up1+up2+up3);
    const ℝ test1 = fabs(dot2D(dxx1,duav)*dx2);
    const ℝ test2 = fabs(dot2D(dxx2,duav)*dx1);
    const ℝ num = (test1>test2?dx1:dx2);
    const ℝ den = (test1>test2?dx2:dx1);
    const ℝ r = num / den;
    //ℝ evol = sqrt(4.0 * cvolume * r);
    const ℝ evol = min(sqrt(4.0*cvolume*r),2.0*minelen);
    // compute delta velocity
    const ℝ dv1 = length2(up1+up2-up0-up3);
    const ℝ dv2 = length2(up2+up3-up0-up1);
    const ℝ du = sqrt(max(dv1, dv2));
    c0evol = (c0div < 0.0 ? evol : 0.);
    //dbc(green,"[QCS::setCornerDiv] c0evol="<<c0evol);
    c0du   = (c0div < 0.0 ? du   : 0.);
    //dbc(green,"[QCS::setCornerDiv] c0du="<<c0du);
  }
}

// [4.1] Compute the c0rmu (real Kurapatenko viscous scalar)
∀ cells qcsCalcForce_setQCnForce_4_1 @ 6.4 {
  const ℝ gammap1 = qgamma + 1.0;
  ∀ node {
    // Kurapatenko form of the viscosity
    const ℝ ztmp2 = q2 * 0.25 * gammap1 * c0du;
    const ℝ ztmp1 = q1 * zss;
    const ℝ zkur = ztmp2 + sqrt(ztmp2 * ztmp2 + ztmp1 * ztmp1);
    // Compute c0rmu for each corner
    const ℝ rmu = zkur * zrp * c0evol;
    c0rmu = ((c0div > 0.0) ? 0. : rmu);
    //dbc(green,"[QCS::setQCnForce] c0rmu="<<c0rmu);
  }
}

// [4.2] Compute the c0qe for each corner
∀ cells qcsCalcForce_setQCnForce_4_2 @ 6.5 {
  const int N = nbNode;
  const ℝ³ _pu[4] = {pu[0],pu[1],pu[2],pu[3]};
  const ℝ _elen[4] = {elen[0],elen[1],elen[2],elen[3]};
  ∀ node {
    const ℕ p1=(#+N-1)%N; // previous
    const ℕ p=(#);        // current
    const ℕ p2=(#+1)%N;   // next
    // Compute: c0qe(1,2,3)=edge 1, y component (2nd), 3rd corner
    //          c0qe(2,1,3)=edge 2, x component (1st)
    //dbc(green,"[QCS::setQCnForce] pu[p]="<<_pu[p]<<", pu[p2]="<<_pu[p2]);
    //dbc(green,"[QCS::setQCnForce] elen[e2]="<<_elen[p]<<", c0rmu="<<c0rmu);
    c0qe0 = c0rmu * (_pu[p] - _pu[p1]) / _elen[p1];
    //dbc(green,"[QCS::setQCnForce] c0qe0="<<c0qe0);
    c0qe1 = c0rmu * (_pu[p2] - _pu[p]) / _elen[p];
    //dbc(green,"[QCS::setQCnForce] c0qe1="<<c0qe1);
  }
}
//∀ cells @ 6.51 { ∀ node dbc(green,"[QCS::setQCnForce] c0qe0="<<c0qe0);}
//∀ cells @ 6.52 { ∀ node dbc(green,"[QCS::setQCnForce] c0qe1="<<c0qe1);}

// [5] Compute the Q forces
// Routine number [5]  in the full algorithm CS2DQforce(...)
// [5.1] Preparation of extra variables
∀ cells qcsCalcForce_setForce_preparationOfExtraVariables @ 6.6 {
  ∀ node {
    const ℝ csin2 = 1.0 - c0cos * c0cos;
    //dbc(green,"[QCS::setForce] csin2="<<csin2);
    c0w   = ((csin2 < 1.e-4) ? 0. : c0area / csin2);
    c0cos = ((csin2 < 1.e-4) ? 0. : c0cos);
    //dbc(green,"[QCS::setForce] c0cos="<<c0cos);
  }
}
//∀ cells @ 6.61 { ∀ node dbc(green,"[QCS::setForce] c0w="<<c0w);}
//∀ cells @ 6.62 { ∀ node dbc(green,"[QCS::setForce] c0cos="<<c0cos);}

∀ cells qcsCalcForce_setForce_setUpTheForcesOnCorners @ 6.7 {
  const ℕ N = nbNode;
  const ℝ _elen[4] = {elen[0],elen[1],elen[2],elen[3]};
  ∀ face {
    // Edge length for c1, c2 contribution to s
    const ℝ _el = _elen[#];
    const ℕ c20 = (#+1)%N; // next
    sfq = (c0w*(c0qe1+c0cos*c0qe0)
           +c0w[NABLA_FACE_PER_CELL*c+c20]
           *(c0qe0[NABLA_FACE_PER_CELL*c+c20]
             +c0cos[NABLA_FACE_PER_CELL*c+c20]
             *c0qe1[NABLA_FACE_PER_CELL*c+c20]))/_el;
    dbc(green,"[QCS::setForce] sfq="<<sfq);
  }
}
// Routine number [6] in the full algorithm
∀ cells qcsCalcForce_setVelDiff @ 6.8 {
  ℝ z0tmp=0;
  const ℕ N = nbNode;
  ℝ³ u[4]; ∀ node u[#]=pu;
  ℝ³ x[4]; ∀ node x[#]=pxp;
  ℝ e[4]; ∀ face e[#]=elen;
  ∀ node {
    //const ℕ i=(#+N-1)%N; // previous
    const ℕ j=(#);       // current
    const ℕ k=(#+1)%N; // next
    const ℝ³ dx = x[k]-x[j];
    const ℝ³ du = u[k]-u[j];
    //dbc(green,"[QCS::setVelDiff] dx="<<dx<<", du="<<du);
    const ℝ lenx = e[j];
    const ℝ dux = (lenx > 0. ? fabs(dot2D(du,dx)) / lenx : 0.);
    //dbc(green,"[QCS::setVelDiff] dux="<<dux);
    z0tmp = max(z0tmp, dux);
    //dbc(green,"[QCS::setVelDiff] z0tmp="<<z0tmp);
  }
  zdu = q1 * zss + 2. * q2 * z0tmp;
  dbc(green,"[QCS::setVelDiff] zdu="<<zdu);
}
//∀ cells @ 6.61 { ∀ node dbc(green,"[QCS::setVelDiff] zdu="<<zdu);}

∀ cells sumCrnrForce @ 7.0 {
  const ℕ N = nbNode;
  ℝ³ p[4]; ∀ face p[#]=sfp;
  ℝ³ q[4]; ∀ face q[#]=sfq;
  ℝ³ t[4]; ∀ face t[#]=sft;
  ∀ node {
    const ℕ i=(#+N-1)%N; // previous
    //dbc(white,"[Hydro::sumCrnrForce] s="<<#<<", s3="<<i);
    //dbc(white,"[Hydro::sumCrnrForce] sf[s]="<<p[#]<<", sf2[s]="<<q[#]<<", sf3[s]="<<t[#]);
    cftot = (p[#] + q[#] + t[#]) - (p[i] + q[i] + t[i]);
    dbc(white,"[QCS::sumCrnrForce] cftot["<<(uid*4+n)<<"]="<<cftot);
  }
}

∀ nodes sumCornerMassesToPoints @ 8.0{
  pmaswt=0;
  ∀ cell pmaswt+=cmaswt;
  dbc(yellow,"[sumCornerMassesToPoints] pmaswt="<<pmaswt);
}

∀ nodes sumCornerForcesToPoints @ 8.1{
  pf=0;
  //dbc(green,"[sumCornerForcesToPoints] node #"<<n);
  ∀ cell {
/*    dbc(green,"[sumCornerForcesToPoints] "<<
      " xs_node_cell="<<xs_node_cell[NABLA_NODE_PER_CELL*(n<<WARP_BIT)+c]<<
      " xs_node_cell_corner="<<xs_node_cell_corner[NABLA_NODE_PER_CELL*(n<<WARP_BIT)+c]<<
      " gathered_cell_cftot="<<gathered_cell_cftot
      );*/
    pf+=cftot;
  }
  dbc(yellow,"[sumCornerForcesToPoints] pf="<<pf);
}
 
// 4a. apply boundary conditions
∀ /*outer*/ nodes applyFixedBC_X @ 9.0 {
  if (px.x!=0.0 && px.x!=LENGTH) continue;
  //dbc(white,"[applyFixedBC] X p="<<n);
  const ℝ³ vfixx = ℝ³(1.,0.,0.);
  pu = project(pu, vfixx);
  pf = project(pf, vfixx);
}
∀ /*outer*/ nodes applyFixedBC_Y @ 9.1 {
  if (px.y!=0.0 && px.y!=LENGTH) continue;
  //dbc(white,"[applyFixedBC] Y p="<<n);
  const ℝ³ vfixy = ℝ³(0.,1.,0.);
  pu = project(pu, vfixy);
  pf = project(pf, vfixy);
} 

∀ nodes calcAccel @ 10.0 {
  const double fuzz = 1.e-99;
  pap = pf / max(pmaswt, fuzz);
  dbc(white,"[calcAccel] pap="<<pap);
}

// ===== Corrector step =====
// 6. advance mesh to end of time step
∀ nodes advPosFull @ 11.0 {
  pu = pu0 + pap * dt;
  //dbc(white,"[Hydro::advPosFull] pu0="<<pu0);
  //dbc(white,"[Hydro::advPosFull] pu="<<pu);
  px = px0 + ½ * (pu + pu0) * dt;
  //dbc(white,"[Hydro::advPosFull] px="<<px);
}

resetDtHydro @ 12.0 { dtrec = 1.e99; }

// 6a. compute new mesh geometry
∀ cells computeNewMeshGeometry @ 13.0 {
  zx=0.0;
  ∀ node zx+=px;
  zx /= nbNode;
  dbc(yellow,"[computeNewMeshGeometry] zx["<<c<<"]="<<zx);
}
∀ faces computeExpBis @ 13.1 {
  ex = 0.0;
  ∀ node ex += px;
  ex /= nbNode;
  assert(nbNode==2);
  //dbc(yellow,"[computeExpBis] ex["<<uid<<"]="<<ex);
}
∀ cells calcVolsBis @ 13.2 { 
  const int N = nbNode;
  ℝ³ X[4]; ∀ node X[#]=px;
  zvol=0;
  zarea=0; 
  ∀ node {
    const ℕ i=#;
    const ℕ j=(#+1)%N;
    sarea = ½ * cross2D(X[j]-X[i],zx-X[i]);
    svol = ⅓* sarea * (X[i].x+X[j].x+zx.x);
    zarea += sarea;
    zvol  += svol;
    //dbc(white,"[calcVolsBis] sarea["<<uid<<","<<n<<"]="<<sarea);
  }
  dbc(yellow,"[calcVolsBis] zvol="<<zvol);
}

// 7. compute work
∀ cells calcWork @ 14.0 { // sfp, sfq, pu0, pu, pxp, dt, zw, zetot,
  zw=0; 
  const int N = nbNode;
  const ℝ dth = ½ * dt;
  //dbc(white,"[calcWork] dth="<<dth);
  ℝ³ _pu[4]; ∀ node _pu[#]=pu;
  ℝ³ _pu0[4]; ∀ node _pu0[#]=pu0;
  const ℝ _xp[4] = {pxp[0].x,pxp[1].x,pxp[2].x,pxp[3].x};
  ∀ face { 
    const ℕ j=(#);
    const ℕ k=(#+1)%N;
    const ℝ³ sftot = sfp + sfq;
    const ℝ sd1 = dot2D( sftot, (_pu0[j] + _pu[j]));
    const ℝ sd2 = dot2D(-sftot, (_pu0[k] + _pu[k]));
    const ℝ dwork = -dth * (sd1 * _xp[j] + sd2 * _xp[k]);
    zetot += dwork;
    zw += dwork;
  }
  dbc(white,"[calcWork] zw="<<zw);
}

// 7a. compute work rate
∀ cells calcWorkRate @ 15.0 { //(zvol0, zvol, zw, zp, dt, zwrate, zfirst, zlast);
  const ℝ dtinv = 1. / dt;
  const ℝ dvol = zvol - zvol0;
  zwrate = (zw + zp * dvol) * dtinv;
  dbc(white,"[calcWorkRate] zwrate="<<zwrate);
}

// 8. update state variables
∀ cells calcEnergy @ 16.0 { //(zetot, zm, ze, zfirst, zlast);
  const ℝ fuzz = 1.e-99;
  ze = zetot / (zm + fuzz);
  dbc(white,"[Hydro::calcEnergy] ze="<<ze);
}

∀ cells calcRho @ 17.0 { //(zm, zvol, zr, zfirst, zlast);
  zr = zm / zvol;
  dbc(white,"[calcRho] zr="<<zr);
}

// 9.  compute timestep for next cycle

∀ cells calcDtCourant @ 19.0 {
  const ℝ fuzz = 1.e-99;
  ℝ dtnew = 1.e99;
  const ℝ cdu = max(zdu, max(zss, fuzz));
  const ℝ zdthyd = zdl * cfl / cdu;
  dtnew = (zdthyd < dtnew ? zdthyd : dtnew);
  //dbc(white,"[Hydro::calcDtCourant] dtnew="<<dtnew);
  if (dtnew < dtrec){
    //dbc(white,"[calcDtCourant] dtrec="<<dtrec);
    dtrec = dtnew;
  }
}
calcDtCourantDump @ 19.1 {
  dbc(white,"[Hydro::calcDtCourant] dtrec="<<dtrec);
}

calcDtVolumeIni @ 20.0 {
  dvovmax = 1.e-99;
}
∀ cells calcDtVolume @ 20.0 {
  const ℝ zdvov = fabs((zvol - zvol0) / zvol0);
  dvovmax = (zdvov > dvovmax ? zdvov : dvovmax);
  dbc(white,"[Hydro::calcDtVolume] dvovmax="<<dvovmax);
}
calcDtVolumeTst @ 20.1 {
  const ℝ dtnew = dt * cflv / dvovmax;
  //dbc(white,"[Hydro::calcDtVolume] dt="<<dt);
  //dbc(white,"[Hydro::calcDtVolume] cflv="<<cflv);
  //dbc(white,"[Hydro::calcDtVolume] dvovmax="<<dvovmax);
  dbc(white,"[Hydro::calcDtVolume] dtnew="<<dtnew);
  if (dtnew < dtrec){
    //dbc(white,"[calcDtVolume] dtrec="<<dtrec);
    dtrec = dtnew;
  }
}

calcDtHydro @ 20.2 {
  dbc(white,"[calcDtHydro] dtrec="<<dtrec);
}

// End of Hydro::doCycle

incTime @ 30.0 {
  time += dt;
  cout<<"[35mEnd cycle "<<cycle
      <<", time = "<<time<<", dt = "<<dt<<"[m"<<endl;
}

tstExit @ 40.0 {
  if (cycle < cstop && time < tstop) return;
  exit;
}
